import click
import eth_typing
from sqlalchemy import select

from degenbot.checksum_cache import get_checksum_address
from degenbot.cli import cli
from degenbot.config import CONFIG_FILE, settings
from degenbot.database import db_session
from degenbot.database.models.base import ExchangeTable
from degenbot.database.models.pools import PoolManagerTable


@cli.group()
def exchange() -> None:
    """
    Exchange commands
    """


@exchange.group
def activate() -> None:
    """
    Activate the exchange. Liquidity pools for all activated exchanges are included when running
    "pool update".
    """


@exchange.group
def deactivate() -> None:
    """
    Deactivate the exchange. Liquidity pools for all deactivated exchanges are not included when
    running "pool update".
    """


def _check_configured_rpc(chain_id: int) -> None:
    if chain_id not in settings.rpc:
        click.echo(
            f"An RPC for chain ID {chain_id} is not defined. Add one to {CONFIG_FILE} so that updates can be performed from the console."  # noqa: E501
        )


@activate.command("base_aerodrome_v2")
def activate_base_aerodrome_v2(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "aerodrome_v2",
) -> None:
    """
    Activate Aerodrome V2 on Base mainnet.
    """

    _check_configured_rpc(chain_id)

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )
    if exchange is not None:
        if exchange.active:
            click.echo("Exchange is already activated.")
            return
        exchange.active = True
        db_session.commit()

    if exchange is None:
        db_session.add(
            ExchangeTable(
                chain_id=chain_id,
                name=exchange_name,
                active=True,
                factory=get_checksum_address("0x420DD381b31aEf6683db6B902084cB0FFECe40Da"),
            )
        )
        db_session.commit()

    click.echo(f"Activated Aerodrome V2 on Base (chain ID {chain_id}).")


@activate.command("base_aerodrome_v3")
def activate_base_aerodrome_v3(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "aerodrome_v3",
) -> None:
    """
    Activate Aerodrome V2 on Base mainnet.
    """

    _check_configured_rpc(chain_id)

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )
    if exchange is not None:
        if exchange.active:
            click.echo("Exchange is already activated.")
            return
        exchange.active = True
        db_session.commit()

    if exchange is None:
        db_session.add(
            ExchangeTable(
                chain_id=chain_id,
                name=exchange_name,
                active=True,
                factory=get_checksum_address("0x5e7BB104d84c7CB9B682AaC2F3d509f5F406809A"),
            )
        )
        db_session.commit()

    click.echo(f"Activated Aerodrome V3 on Base (chain ID {chain_id}).")


@activate.command("base_pancakeswap_v2")
def activate_base_pancakeswap_v2(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "pancakeswap_v2",
) -> None:
    """
    Activate Pancakeswap V2 on Base mainnet.
    """

    _check_configured_rpc(chain_id)

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )
    if exchange is not None:
        if exchange.active:
            click.echo("Exchange is already activated.")
            return
        exchange.active = True
        db_session.commit()

    if exchange is None:
        db_session.add(
            ExchangeTable(
                chain_id=chain_id,
                name=exchange_name,
                active=True,
                factory=get_checksum_address("0x02a84c1b3BBD7401a5f7fa98a384EBC70bB5749E"),
            )
        )
        db_session.commit()

    click.echo(f"Activated Pancakeswap V2 on Base (chain ID {chain_id}).")


@activate.command("base_pancakeswap_v3")
def activate_base_pancakeswap_v3(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "pancakeswap_v3",
) -> None:
    """
    Activate Pancakeswap V3 on Base mainnet.
    """

    _check_configured_rpc(chain_id)

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )
    if exchange is not None:
        if exchange.active:
            click.echo("Exchange is already activated.")
            return
        exchange.active = True
        db_session.commit()

    if exchange is None:
        db_session.add(
            ExchangeTable(
                chain_id=chain_id,
                name=exchange_name,
                active=True,
                factory=get_checksum_address("0x0BFbCF9fa4f9C56B0F40a671Ad40E0805A091865"),
                deployer=get_checksum_address("0x41ff9AA7e16B8B1a8a8dc4f0eFacd93D02d071c9"),
            )
        )
        db_session.commit()

    click.echo(f"Activated Pancakeswap V3 on Base (chain ID {chain_id}).")


@activate.command("base_swapbased_v2")
def activate_base_swapbased_v2(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "swapbased_v2",
) -> None:
    """
    Activate SwapBased V2 on Base mainnet.
    """

    _check_configured_rpc(chain_id)

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )
    if exchange is not None:
        if exchange.active:
            click.echo("Exchange is already activated.")
            return
        exchange.active = True
        db_session.commit()

    if exchange is None:
        db_session.add(
            ExchangeTable(
                chain_id=chain_id,
                name=exchange_name,
                active=True,
                factory=get_checksum_address("0x04C9f118d21e8B767D2e50C946f0cC9F6C367300"),
            )
        )
        db_session.commit()

    click.echo(f"Activated SwapBased V2 on Base (chain ID {chain_id}).")


@activate.command("base_sushiswap_v2")
def activate_base_sushiswap_v2(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "sushiswap_v2",
) -> None:
    """
    Activate Sushiswap V2 on Base mainnet.
    """

    _check_configured_rpc(chain_id)

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )
    if exchange is not None:
        if exchange.active:
            click.echo("Exchange is already activated.")
            return
        exchange.active = True
        db_session.commit()

    if exchange is None:
        db_session.add(
            ExchangeTable(
                chain_id=chain_id,
                name=exchange_name,
                active=True,
                factory=get_checksum_address("0x71524B4f93c58fcbF659783284E38825f0622859"),
            )
        )
        db_session.commit()

    click.echo(f"Activated Sushiswap V2 on Base (chain ID {chain_id}).")


@activate.command("base_sushiswap_v3")
def activate_base_sushiswap_v3(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "sushiswap_v3",
) -> None:
    """
    Activate Sushiswap V3 on Base mainnet.
    """

    _check_configured_rpc(chain_id)

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )
    if exchange is not None:
        if exchange.active:
            click.echo("Exchange is already activated.")
            return
        exchange.active = True
        db_session.commit()

    if exchange is None:
        db_session.add(
            ExchangeTable(
                chain_id=chain_id,
                name=exchange_name,
                active=True,
                factory=get_checksum_address("0xc35DADB65012eC5796536bD9864eD8773aBc74C4"),
            )
        )
        db_session.commit()

    click.echo(f"Activated Sushiswap V3 on Base (chain ID {chain_id}).")


@activate.command("base_uniswap_v2")
def activate_base_uniswap_v2(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "uniswap_v2",
) -> None:
    """
    Activate Uniswap V2 on Base mainnet.
    """

    _check_configured_rpc(chain_id)

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )
    if exchange is not None:
        if exchange.active:
            click.echo("Exchange is already activated.")
            return
        exchange.active = True
        db_session.commit()

    if exchange is None:
        db_session.add(
            ExchangeTable(
                chain_id=chain_id,
                name=exchange_name,
                active=True,
                factory=get_checksum_address("0x8909Dc15e40173Ff4699343b6eB8132c65e18eC6"),
            )
        )
        db_session.commit()

    click.echo(f"Activated Uniswap V2 on Base (chain ID {chain_id}).")


@activate.command("base_uniswap_v3")
def activate_base_uniswap_v3(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "uniswap_v3",
) -> None:
    """
    Activate Uniswap V3 on Base mainnet.
    """

    _check_configured_rpc(chain_id)

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )
    if exchange is not None:
        if exchange.active:
            click.echo("Exchange is already activated.")
            return
        exchange.active = True
        db_session.commit()

    if exchange is None:
        db_session.add(
            ExchangeTable(
                chain_id=chain_id,
                name=exchange_name,
                active=True,
                factory=get_checksum_address("0x33128a8fC17869897dcE68Ed026d694621f6FDfD"),
            )
        )
        db_session.commit()

    click.echo(f"Activated Uniswap V3 on Base (chain ID {chain_id}).")


@activate.command("base_uniswap_v4")
def activate_base_uniswap_v4(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "uniswap_v4",
) -> None:
    """
    Activate Uniswap V4 on Base mainnet.
    """

    _check_configured_rpc(chain_id)

    pool_manager_address = get_checksum_address("0x498581fF718922c3f8e6A244956aF099B2652b2b")
    exchange_kind = "uniswap_v4"

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )
    if exchange is not None:
        if exchange.active:
            click.echo("Exchange is already activated.")
            return
        exchange.active = True
        db_session.commit()

    if exchange is None:
        exchange = ExchangeTable(
            chain_id=chain_id,
            name=exchange_name,
            active=True,
            factory=pool_manager_address,
        )
        db_session.add(exchange)
        db_session.flush()

        manager_in_db = db_session.scalar(
            select(PoolManagerTable).where(PoolManagerTable.address == pool_manager_address)
        )
        if manager_in_db is None:
            db_session.add(
                PoolManagerTable(
                    address=pool_manager_address,
                    chain=chain_id,
                    kind=exchange_kind,
                    exchange_id=exchange.id,
                )
            )

        db_session.commit()

    click.echo(f"Activated Uniswap V4 on Base (chain ID {chain_id}).")


@deactivate.command("base_aerodrome_v2")
def deactivate_base_aerodrome_v2(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "aerodrome_v2",
) -> None:
    """
    Deactivate Aerodrome V2 on Base mainnet.
    """

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )

    if exchange is None:
        click.echo(f"The database has no entry for Aerodrome V2 on Base (chain ID {chain_id}).")
        return

    if not exchange.active:
        click.echo("Exchange is already deactivated.")
        return
    exchange.active = False
    db_session.commit()

    click.echo(f"Deactivated Aerodrome V2 on Base (chain ID {chain_id}).")


@deactivate.command("base_aerodrome_v3")
def deactivate_base_aerodrome_v3(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "aerodrome_v3",
) -> None:
    """
    Deactivate Aerodrome V3 on Base mainnet.
    """

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )

    if exchange is None:
        click.echo(f"The database has no entry for Aerodrome V3 on Base (chain ID {chain_id}).")
        return

    if not exchange.active:
        click.echo("Exchange is already deactivated.")
        return
    exchange.active = False
    db_session.commit()

    click.echo(f"Deactivated Aerodrome V3 on Base (chain ID {chain_id}).")


@deactivate.command("base_pancakeswap_v2")
def deactivate_base_pancakeswap_v2(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "pancakeswap_v2",
) -> None:
    """
    Deactivate Pancakeswap V2 on Base mainnet.
    """

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )

    if exchange is None:
        click.echo(f"The database has no entry for Pancakeswap V2 on Base (chain ID {chain_id}).")
        return

    if not exchange.active:
        click.echo("Exchange is already deactivated.")
        return
    exchange.active = False
    db_session.commit()

    click.echo(f"Deactivated Pancakeswap V2 on Base (chain ID {chain_id}).")


@deactivate.command("base_pancakeswap_v3")
def deactivate_base_pancakeswap_v3(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "pancakeswap_v3",
) -> None:
    """
    Deactivate Pancakeswap V3 on Base mainnet.
    """

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )

    if exchange is None:
        click.echo(f"The database has no entry for Pancakeswap V3 on Base (chain ID {chain_id}).")
        return

    if not exchange.active:
        click.echo("Exchange is already deactivated.")
        return
    exchange.active = False
    db_session.commit()

    click.echo(f"Deactivated Pancakeswap V3 on Base (chain ID {chain_id}).")


@deactivate.command("base_sushiswap_v2")
def deactivate_base_sushiswap_v2(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "sushiswap_v2",
) -> None:
    """
    Deactivate Sushiswap V2 on Base mainnet.
    """

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )

    if exchange is None:
        click.echo(f"The database has no entry for Sushiswap V2 on Base (chain ID {chain_id}).")
        return

    if not exchange.active:
        click.echo("Exchange is already deactivated.")
        return
    exchange.active = False
    db_session.commit()

    click.echo(f"Deactivated Sushiswap V2 on Base (chain ID {chain_id}).")


@deactivate.command("base_sushiswap_v3")
def deactivate_base_sushiswap_v3(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "sushiswap_v3",
) -> None:
    """
    Deactivate Sushiswap V3 on Base mainnet.
    """

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )

    if exchange is None:
        click.echo(f"The database has no entry for Sushiswap V3 on Base (chain ID {chain_id}).")
        return

    if not exchange.active:
        click.echo("Exchange is already deactivated.")
        return
    exchange.active = False
    db_session.commit()

    click.echo(f"Deactivated Sushiswap V3 on Base (chain ID {chain_id}).")


@deactivate.command("base_swapbased_v2")
def deactivate_base_swapbased_v2(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "swapbased_v2",
) -> None:
    """
    Deactivate SwapBased V2 on Base mainnet.
    """

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )

    if exchange is None:
        click.echo(f"The database has no entry for SwapBased V2 on Base (chain ID {chain_id}).")
        return

    if not exchange.active:
        click.echo("Exchange is already deactivated.")
        return
    exchange.active = False
    db_session.commit()

    click.echo(f"Deactivated SwapBased V2 on Base (chain ID {chain_id}).")


@deactivate.command("base_uniswap_v2")
def deactivate_base_uniswap_v2(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "uniswap_v2",
) -> None:
    """
    Deactivate Uniswap V2 on Base mainnet.
    """

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )

    if exchange is None:
        click.echo(f"The database has no entry for Uniswap V2 on Base (chain ID {chain_id}).")
        return

    if not exchange.active:
        click.echo("Exchange is already deactivated.")
        return
    exchange.active = False
    db_session.commit()

    click.echo(f"Deactivated Uniswap V2 on Base (chain ID {chain_id}).")


@deactivate.command("base_uniswap_v3")
def deactivate_base_uniswap_v3(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "uniswap_v3",
) -> None:
    """
    Deactivate Uniswap V3 on Base mainnet.
    """

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )

    if exchange is None:
        click.echo(f"The database has no entry for Uniswap V3 on Base (chain ID {chain_id}).")
        return

    if not exchange.active:
        click.echo("Exchange is already deactivated.")
        return
    exchange.active = False
    db_session.commit()

    click.echo(f"Deactivated Uniswap V3 on Base (chain ID {chain_id}).")


@deactivate.command("base_uniswap_v4")
def deactivate_base_uniswap_v4(
    chain_id: eth_typing.ChainId = eth_typing.ChainId.BASE,
    exchange_name: str = "uniswap_v4",
) -> None:
    """
    Deactivate Uniswap V4 on Base mainnet.
    """

    exchange = db_session.scalar(
        select(ExchangeTable).where(
            ExchangeTable.chain_id == chain_id,
            ExchangeTable.name == exchange_name,
        )
    )

    if exchange is None:
        click.echo(f"The database has no entry for Uniswap V4 on Base (chain ID {chain_id}).")
        return

    if not exchange.active:
        click.echo("Exchange is already deactivated.")
        return
    exchange.active = False
    db_session.commit()

    click.echo(f"Deactivated Uniswap V4 on Base (chain ID {chain_id}).")



"""add Uniswap V4

Revision ID: 756fba1f75f4
Revises: 9347bbfcd47a
Create Date: 2025-08-04 20:56:04.525682

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

from degenbot.database.models import IntMappedToString

# revision identifiers, used by Alembic.
revision: str = "756fba1f75f4"
down_revision: str | Sequence[str] | None = "9347bbfcd47a"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "pool_managers",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("address", sa.String(length=42), nullable=False),
        sa.Column("chain", sa.Integer(), nullable=False),
        sa.Column("kind", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_pool_managers_address"), "pool_managers", ["address"], unique=False)
    op.create_table(
        "managed_pools",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("manager_id", sa.Integer(), nullable=False),
        sa.Column("kind", sa.Text(), nullable=False),
        sa.ForeignKeyConstraint(
            ["manager_id"],
            ["pool_managers.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "managed_pool_initialization_maps",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("managed_pool_id", sa.Integer(), nullable=False),
        sa.Column("word", sa.Integer(), nullable=False),
        sa.Column("bitmap", IntMappedToString(length=78), nullable=False),
        sa.ForeignKeyConstraint(
            ["managed_pool_id"],
            ["managed_pools.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "ix_managed_pool_initialization_maps_pool_id_word",
        "managed_pool_initialization_maps",
        ["managed_pool_id", "word"],
        unique=True,
    )
    op.create_table(
        "managed_pool_liquidity_positions",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("managed_pool_id", sa.Integer(), nullable=False),
        sa.Column("tick", sa.Integer(), nullable=False),
        sa.Column("liquidity_net", IntMappedToString(length=78), nullable=False),
        sa.Column("liquidity_gross", IntMappedToString(length=78), nullable=False),
        sa.ForeignKeyConstraint(
            ["managed_pool_id"],
            ["managed_pools.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "ix_managed_pool_liquidity_positions_pool_id_tick",
        "managed_pool_liquidity_positions",
        ["managed_pool_id", "tick"],
        unique=True,
    )
    op.create_table(
        "uniswap_v4_pools",
        sa.Column("managed_pool_id", sa.Integer(), nullable=False),
        sa.Column("pool_hash", sa.LargeBinary(), nullable=False),
        sa.Column("hooks", sa.String(length=42), nullable=False),
        sa.Column("currency0", sa.String(length=42), nullable=False),
        sa.Column("currency1", sa.String(length=42), nullable=False),
        sa.Column("fee_currency0", sa.Integer(), nullable=False),
        sa.Column("fee_currency1", sa.Integer(), nullable=False),
        sa.Column("fee_denominator", sa.Integer(), nullable=False),
        sa.Column("tick_spacing", sa.Integer(), nullable=False),
        sa.Column("has_liquidity", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["managed_pool_id"],
            ["managed_pools.id"],
        ),
        sa.PrimaryKeyConstraint("managed_pool_id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("uniswap_v4_pools")
    op.drop_index(
        "ix_managed_pool_liquidity_positions_pool_id_tick",
        table_name="managed_pool_liquidity_positions",
    )
    op.drop_table("managed_pool_liquidity_positions")
    op.drop_index(
        "ix_managed_pool_initialization_maps_pool_id_word",
        table_name="managed_pool_initialization_maps",
    )
    op.drop_table("managed_pool_initialization_maps")
    op.drop_table("managed_pools")
    op.drop_index(op.f("ix_pool_managers_address"), table_name="pool_managers")
    op.drop_table("pool_managers")
    # ### end Alembic commands ###



"""add exchange relationship

Revision ID: 4eada4ae4a55
Revises: 7dc2ca38053f
Create Date: 2025-10-16 14:01:57.396204

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "4eada4ae4a55"
down_revision: str | Sequence[str] | None = "7dc2ca38053f"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    with op.batch_alter_table("pools", schema=None) as batch_op:
        batch_op.add_column(sa.Column("exchange_id", sa.Integer(), nullable=True))
        batch_op.create_foreign_key(
            "fk_pools_exchange_id_exchanges",
            "exchanges",
            ["exchange_id"],
            ["id"],
        )

    connection = op.get_bind()
    rows = connection.execute(sa.text("SELECT id, name, chain_id FROM exchanges")).fetchall()
    exchange_map = {
        (exchange_name, exchange_chain_id): exchange_id
        for (
            exchange_id,
            exchange_name,
            exchange_chain_id,
        ) in rows
    }

    for (name, chain_id), exchange_id in exchange_map.items():
        connection.execute(
            sa.text(
                """
                UPDATE pools
                SET exchange_id = :exchange_id
                WHERE kind = :name AND chain = :chain_id
                """
            ),
            {
                "exchange_id": exchange_id,
                "name": name,
                "chain_id": chain_id,
            },
        )

    # Step 3: Make column non-nullable after data filled
    with op.batch_alter_table("pools", schema=None) as batch_op:
        batch_op.alter_column("exchange_id", existing_type=sa.Integer(), nullable=False)


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



def div_rounding_up(
    x: int,
    y: int,
) -> int:
    """
    Calculates ceil(x/y)

    @dev division by 0 will return 0, and should be checked externally.
    """

    return 0 if y == 0 else x // y + int(x % y > 0)


def simple_mul_div(a: int, b: int, denominator: int) -> int:
    """
    Calculates floor((a*b)/denominator))

    @dev division by 0 will return 0, and should be checked externally
    """

    return 0 if denominator == 0 else (a * b) // denominator



"""drop factory column

Revision ID: 04f858f979a9
Revises: 8b29ed99a2ac
Create Date: 2025-10-18 00:07:36.604383

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "04f858f979a9"
down_revision: str | Sequence[str] | None = "8b29ed99a2ac"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""

    op.drop_table("metadata")
    with op.batch_alter_table("aerodrome_v2_pools", schema=None) as batch_op:
        batch_op.drop_column("factory")

    with op.batch_alter_table("aerodrome_v3_pools", schema=None) as batch_op:
        batch_op.drop_column("factory")

    with op.batch_alter_table("camelot_v2_pools", schema=None) as batch_op:
        batch_op.drop_column("factory")

    with op.batch_alter_table("pancakeswap_v2_pools", schema=None) as batch_op:
        batch_op.drop_column("factory")

    with op.batch_alter_table("pancakeswap_v3_pools", schema=None) as batch_op:
        batch_op.drop_column("factory")

    with op.batch_alter_table("sushiswap_v2_pools", schema=None) as batch_op:
        batch_op.drop_column("factory")

    with op.batch_alter_table("sushiswap_v3_pools", schema=None) as batch_op:
        batch_op.drop_column("factory")

    with op.batch_alter_table("swapbased_v2_pools", schema=None) as batch_op:
        batch_op.drop_column("factory")

    with op.batch_alter_table("uniswap_v2_pools", schema=None) as batch_op:
        batch_op.drop_column("factory")

    with op.batch_alter_table("uniswap_v3_pools", schema=None) as batch_op:
        batch_op.drop_column("factory")


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



from . import (
    abi as abi,
)  # excluded from __all__ so it doesn't bubble back up to the top level package namespace
from .curve_stableswap_liquidity_pool import CurveStableswapPool
from .types import (
    CurveStableswapPoolSimulationResult,
    CurveStableswapPoolState,
    CurveStableSwapPoolStateUpdated,
)

__all__ = (
    "CurveStableSwapPoolStateUpdated",
    "CurveStableswapPool",
    "CurveStableswapPoolSimulationResult",
    "CurveStableswapPoolState",
)



"""Rename pool hash column, add unique constraint

Revision ID: a4f59783919f
Revises: 3199199def8c
Create Date: 2025-09-05 13:47:09.094614

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "a4f59783919f"
down_revision: str | Sequence[str] | None = "3199199def8c"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    with op.batch_alter_table("uniswap_v4_pools") as batch_op:
        batch_op.alter_column("pool_hash_hex", new_column_name="pool_hash")


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



from degenbot.constants import MAX_UINT128, MIN_UINT128
from degenbot.exceptions.evm import EVMRevertError


def add_delta(
    x: int,
    y: int,
) -> int:
    """
    This function has been modified to check that the result fits in a uint128, instead
    of inline Yul as implemented by the Solidity contract.

    ref: https://github.com/Uniswap/v4-core/blob/main/src/libraries/LiquidityMath.sol
    """

    result = x + y
    if result < MIN_UINT128 or result > MAX_UINT128:
        msg = "SafeCastOverflow"
        raise EVMRevertError(msg)

    return result



import asyncio
import pathlib
from collections import defaultdict, deque
from typing import TYPE_CHECKING, Any, Protocol, TypedDict

import pydantic_core
import tqdm
import tqdm.asyncio
from eth_abi.abi import decode as abi_decode
from eth_typing import ChecksumAddress, HexAddress
from hexbytes import HexBytes
from sqlalchemy import select
from web3 import Web3
from web3.types import LogReceipt

from degenbot.checksum_cache import get_checksum_address
from degenbot.config import settings
from degenbot.connection import async_connection_manager, connection_manager
from degenbot.database import db_session
from degenbot.database.models.base import ExchangeTable
from degenbot.database.models.pools import AbstractUniswapV3Pool, LiquidityPoolTable
from degenbot.database.operations import get_scoped_sqlite_session
from degenbot.exceptions.liquidity_pool import UnknownPool
from degenbot.functions import fetch_logs_retrying, fetch_logs_retrying_async
from degenbot.logging import logger
from degenbot.types.aliases import BlockNumber, ChainId
from degenbot.types.concrete import KeyedDefaultDict
from degenbot.uniswap.abi import UNISWAP_V3_POOL_ABI
from degenbot.uniswap.v3_types import (
    UniswapV3BitmapAtWord,
    UniswapV3LiquidityAtTick,
    UniswapV3LiquidityEvent,
    UniswapV3PoolLiquidityMappingUpdate,
)

if TYPE_CHECKING:
    from collections.abc import Sequence


class LiquidityMap(TypedDict):
    tick_bitmap: dict[int, UniswapV3BitmapAtWord]
    tick_data: dict[int, UniswapV3LiquidityAtTick]


class UniswapV3LiquiditySnapshotSource(Protocol):
    """
    A minimal protocol allowing the UniswapV3LiquiditySnapshot class to retrieve pool data from a
    generic source.
    """

    storage_kind: str
    chain_id: int

    # Any class implementing the protocol must implement these methods, transforming data as
    # necessary to return the specified types.
    def get_liquidity_map(self, pool_address: ChecksumAddress) -> LiquidityMap | None: ...
    def get_newest_block(self) -> BlockNumber | None: ...
    def get_pools(self) -> set[ChecksumAddress]: ...


class MonolithicJsonFileSnapshot:
    """
    A pool liquidity source backed by a single JSON file with this structure:
    {
        "snapshot_block": int,
        "chain_id": int,
        "0xPoolAddress1": {
            "tick_bitmap": {
                <word>: {
                    'bitmap': <value>,
                    'block': <value>,
                },
                ...
            },
            "tick_data": {
                <tick>: {
                    'liquidity_gross: <value>,
                    'liquidity_net': <value>,
                    'block: <value>,
                }
            }
        },
        "0xPoolAddress2": { ... },
        "0xPoolAddress3": { ... },
        ...
    }
    """

    storage_kind = "file"

    def __init__(self, path: pathlib.Path | str) -> None:
        path = pathlib.Path(path).expanduser().absolute()
        self._path = path
        self._file_snapshot: dict[str, Any] = pydantic_core.from_json(path.read_bytes())
        self.chain_id: int = self._file_snapshot["chain_id"]

    def get_liquidity_map(self, pool_address: ChecksumAddress) -> LiquidityMap | None:
        if pool_address not in self._file_snapshot:
            return None

        return LiquidityMap(
            tick_bitmap={
                int(k): UniswapV3BitmapAtWord(**v)
                for k, v in self._file_snapshot[pool_address]["tick_bitmap"].items()
            },
            tick_data={
                int(k): UniswapV3LiquidityAtTick(**v)
                for k, v in self._file_snapshot[pool_address]["tick_data"].items()
            },
        )

    def get_newest_block(self) -> BlockNumber | None:
        newest_block = self._file_snapshot.get("snapshot_block")
        if newest_block is None:
            return None
        return int(newest_block)

    def get_pools(self) -> set[ChecksumAddress]:
        # all top-level keys except metadata entries
        return {
            get_checksum_address(key)
            for key in self._file_snapshot
            if key not in ("chain_id", "snapshot_block")
        }


class IndividualJsonFileSnapshot:
    """
    Snapshot source backed by a directory of JSON files with this tree structure:

        /path/to/snapshots/
        ├── _metadata.json              -> { "block": int, "chain_id": int }
        ├── 0xPoolAddress1.json         -> { "tick_bitmap": {...}, "tick_data": {...} }
        ├── 0xPoolAddress2.json         -> { "tick_bitmap": {...}, "tick_data": {...} }
        └── 0xPoolAddress3.json         -> { "tick_bitmap": {...}, "tick_data": {...} }

    Each pool file contains the same structure as the monolithic snapshot's per-pool entries.
    """

    storage_kind = "dir"

    def __init__(self, path: pathlib.Path | str) -> None:
        dir_path = pathlib.Path(path).expanduser().absolute()
        assert dir_path.exists()
        assert dir_path.is_dir()
        self._dir = dir_path

        metadata_path = self._dir / "_metadata.json"
        self._metadata: dict[str, Any] = pydantic_core.from_json(metadata_path.read_bytes())
        self.chain_id: int = self._metadata["chain_id"]

    def get_newest_block(self) -> BlockNumber | None:
        newest_block = self._metadata.get("block")
        if newest_block is None:
            return None
        return int(newest_block)

    def get_pools(self) -> set[ChecksumAddress]:
        return {get_checksum_address(pool_file.stem) for pool_file in self._dir.glob("0x*.json")}

    def get_liquidity_map(self, pool_address: ChecksumAddress) -> LiquidityMap | None:
        pool_path = self._dir / f"{pool_address}.json"
        if not pool_path.exists():
            return None

        pool_liquidity_snapshot = pydantic_core.from_json(pool_path.read_bytes())
        return LiquidityMap(
            tick_bitmap={
                int(k): UniswapV3BitmapAtWord(**v)
                for k, v in pool_liquidity_snapshot["tick_bitmap"].items()
            },
            tick_data={
                int(k): UniswapV3LiquidityAtTick(**v)
                for k, v in pool_liquidity_snapshot["tick_data"].items()
            },
        )


class DatabaseSnapshot:
    """
    Snapshot source backed by built-in SQLite database using the ORM abstractions defined
    in `degenbot.database`.

    If a path to an SQLite database is not provided, the default location specified in
    `degenbot.settings` will be used.
    """

    storage_kind = "db"

    def __init__(self, chain_id: ChainId, database_path: pathlib.Path | None = None) -> None:
        if database_path is None:
            self.session = db_session
            self.database_path = settings.database.path
        else:
            self.session = get_scoped_sqlite_session(database_path)
            self.database_path = database_path

        self.chain_id = chain_id

    def get_liquidity_map(self, pool_address: ChecksumAddress) -> LiquidityMap | None:
        pool_in_db = self.session.scalar(
            select(LiquidityPoolTable).where(LiquidityPoolTable.address == pool_address)
        )
        if pool_in_db is None:
            return None

        if TYPE_CHECKING:
            assert isinstance(pool_in_db, AbstractUniswapV3Pool)

        return LiquidityMap(
            tick_bitmap={
                int(initialization_map.word): UniswapV3BitmapAtWord(
                    bitmap=initialization_map.bitmap
                )
                for initialization_map in pool_in_db.initialization_maps
            },
            tick_data={
                int(liquidity_position.tick): UniswapV3LiquidityAtTick(
                    liquidity_gross=liquidity_position.liquidity_gross,
                    liquidity_net=liquidity_position.liquidity_net,
                )
                for liquidity_position in pool_in_db.liquidity_positions
            },
        )

    def get_newest_block(self) -> BlockNumber | None:
        last_update_blocks: Sequence[int | None] = db_session.scalars(
            select(ExchangeTable.last_update_block).where(
                ExchangeTable.chain_id == self.chain_id,
                ExchangeTable.name.like("%!_v3", escape="!"),
            )
        ).all()

        if not last_update_blocks or None in last_update_blocks:
            return None

        return max(
            last_update_block
            for last_update_block in last_update_blocks
            if isinstance(last_update_block, int)
        )

    def get_pools(self) -> set[ChecksumAddress]:
        return {
            get_checksum_address(pool)
            for pool in self.session.scalars(select(AbstractUniswapV3Pool.address)).all()
        }


class UniswapV3LiquiditySnapshot:
    """
    Retrieve and maintain liquidity positions for Uniswap V3 pools.
    """

    UNISWAP_V3_MINT_EVENT_HASH = HexBytes(
        Web3().eth.contract(abi=UNISWAP_V3_POOL_ABI).events.Mint().topic
    )
    UNISWAP_V3_BURN_EVENT_HASH = HexBytes(
        Web3().eth.contract(abi=UNISWAP_V3_POOL_ABI).events.Burn().topic
    )

    def __init__(self, source: UniswapV3LiquiditySnapshotSource) -> None:
        self._source = source
        self._chain_id = source.chain_id

        if (source_block := source.get_newest_block()) is None:
            msg = "The provided source is uninitialized."
            raise ValueError(msg)
        self.newest_block: BlockNumber = source_block

        self._liquidity_events: dict[ChecksumAddress, deque[UniswapV3LiquidityEvent]] = defaultdict(
            deque
        )
        self._liquidity_snapshot: dict[ChecksumAddress, LiquidityMap | None] = KeyedDefaultDict(
            lambda key: self._source.get_liquidity_map(get_checksum_address(key))
        )

        logger.info(f"Loaded Uniswap V3 LP snapshot from {source.storage_kind} source")

    @property
    def chain_id(self) -> int:
        return self._chain_id

    @property
    def pools(self) -> set[ChecksumAddress]:
        return self._source.get_pools()

    def _process_liquidity_event_log(
        self,
        log: LogReceipt,
    ) -> tuple[ChecksumAddress, UniswapV3LiquidityEvent]:
        """
        Decode an event log and convert to an address and a `UniswapV3LiquidityEvent` for
        processing with `UniswapV3Pool.update_liquidity_map`.
        """

        # ref: https://github.com/Uniswap/v3-core/blob/main/contracts/interfaces/pool/IUniswapV3PoolEvents.sol
        #
        # event Mint(
        #     address sender,
        #     address indexed owner,
        #     int24 indexed tickLower,
        #     int24 indexed tickUpper,
        #     uint128 amount,
        #     uint256 amount0,
        #     uint256 amount1
        # );
        # event Burn(
        #     address indexed owner,
        #     int24 indexed tickLower,
        #     int24 indexed tickUpper,
        #     uint128 amount,
        #     uint256 amount0,
        #     uint256 amount1
        # );

        assert not log["removed"]

        (tick_lower,) = abi_decode(["int24"], log["topics"][2])
        (tick_upper,) = abi_decode(["int24"], log["topics"][3])

        if log["topics"][0] == self.UNISWAP_V3_BURN_EVENT_HASH:
            # Decode Burn event
            amount, _, _ = abi_decode(
                ["uint128", "uint256", "uint256"],
                log["data"],
            )
            amount = -amount
        else:
            # Decode Mint event
            _, amount, _, _ = abi_decode(
                ["address", "uint128", "uint256", "uint256"],
                log["data"],
            )

        return log["address"], UniswapV3LiquidityEvent(
            block_number=log["blockNumber"],
            liquidity=amount,
            tick_lower=tick_lower,
            tick_upper=tick_upper,
            tx_index=log["transactionIndex"],
            log_index=log["logIndex"],
        )

    def fetch_new_events(
        self,
        to_block: BlockNumber,
        blocks_per_request: int | None = None,
    ) -> None:
        """
        Fetch liquidity events from the block following the last-known event to the target block
        using `eth_getLogs`. Blocks per request will be capped at `blocks_per_request`.
        """

        logger.info(f"Updating Uniswap V3 snapshot from block {self.newest_block} to {to_block}")

        event_logs = fetch_logs_retrying(
            w3=connection_manager.get_web3(self.chain_id),
            start_block=self.newest_block + 1,
            end_block=to_block,
            max_blocks_per_request=blocks_per_request,
            topic_signature=[
                [
                    self.UNISWAP_V3_MINT_EVENT_HASH,
                    self.UNISWAP_V3_BURN_EVENT_HASH,
                ],  # match topic0: Mint OR Burn
            ],
        )

        for event_log in tqdm.tqdm(
            event_logs,
            desc="Processing liquidity events",
            unit="event",
            bar_format="{desc}: {percentage:3.1f}% |{bar}| {n_fmt}/{total_fmt}",
            leave=False,
        ):
            pool_address, liquidity_event = self._process_liquidity_event_log(event_log)
            self._liquidity_events[pool_address].append(liquidity_event)

        self.newest_block = to_block

    async def fetch_new_events_async(
        self,
        to_block: BlockNumber,
        blocks_per_request: int | None = None,
    ) -> None:
        """
        Async version of fetch_new_events.

        Fetch liquidity events from the block following the last-known event to the target block
        using `eth_getLogs` via the async provider. Blocks per request will be capped at
        `blocks_per_request`.
        """

        logger.info(f"Updating Uniswap V3 snapshot from block {self.newest_block} to {to_block}")

        event_logs = await fetch_logs_retrying_async(
            w3=async_connection_manager.get_web3(self.chain_id),
            start_block=self.newest_block + 1,
            end_block=to_block,
            max_blocks_per_request=blocks_per_request,
            topic_signature=[
                [
                    self.UNISWAP_V3_MINT_EVENT_HASH,
                    self.UNISWAP_V3_BURN_EVENT_HASH,
                ]
            ],
        )

        async for event_log in tqdm.asyncio.tqdm(
            event_logs,
            desc="Processing liquidity events",
            unit="event",
            bar_format="{desc}: {percentage:3.1f}% |{bar}| {n_fmt}/{total_fmt}",
            leave=False,
        ):
            await asyncio.sleep(0)
            pool_address, liquidity_event = self._process_liquidity_event_log(event_log)
            self._liquidity_events[pool_address].append(liquidity_event)

        self.newest_block = to_block

    def pending_updates(
        self,
        pool_address: HexAddress,
    ) -> tuple[UniswapV3PoolLiquidityMappingUpdate, ...]:
        """
        Consume pending liquidity updates for the pool.
        """

        pool_key = get_checksum_address(pool_address)

        try:
            return tuple(
                UniswapV3PoolLiquidityMappingUpdate(
                    block_number=event.block_number,
                    liquidity=event.liquidity,
                    tick_lower=event.tick_lower,
                    tick_upper=event.tick_upper,
                )
                for event in self._liquidity_events[pool_key]
            )
        finally:
            self._liquidity_events[pool_key].clear()

    def tick_bitmap(self, pool_address: str | bytes) -> dict[int, UniswapV3BitmapAtWord] | None:
        """
        Consume the tick initialization bitmaps for the pool.
        """
        pool_address = get_checksum_address(pool_address)

        pool_snapshot = self._liquidity_snapshot[pool_address]
        if pool_snapshot is None:
            return None

        tick_bitmap = pool_snapshot["tick_bitmap"].copy()
        pool_snapshot["tick_bitmap"] = {}
        return tick_bitmap

    def tick_data(self, pool_address: str | bytes) -> dict[int, UniswapV3LiquidityAtTick] | None:
        """
        Consume the liquidity mapping for the pool.
        """
        pool_address = get_checksum_address(pool_address)

        pool_snapshot = self._liquidity_snapshot[pool_address]
        if pool_snapshot is None:
            return None

        tick_data = pool_snapshot["tick_data"].copy()
        pool_snapshot["tick_data"] = {}
        return tick_data

    def update(
        self,
        pool: HexAddress,
        tick_data: dict[int, UniswapV3LiquidityAtTick],
        tick_bitmap: dict[int, UniswapV3BitmapAtWord],
    ) -> None:
        """
        Update the liquidity mapping for the pool.
        """

        pool_key = get_checksum_address(pool)

        pool_snapshot = self._liquidity_snapshot[pool_key]
        if pool_snapshot is None:
            raise UnknownPool(pool_key)

        pool_snapshot["tick_bitmap"].update(tick_bitmap)
        pool_snapshot["tick_data"].update(tick_data)



"""Rename column

Revision ID: 6a77c4e07151
Revises: 082ee8a3d339
Create Date: 2025-10-25 13:49:31.086818

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "6a77c4e07151"
down_revision: str | Sequence[str] | None = "082ee8a3d339"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""

    with op.batch_alter_table("pools", recreate="always", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_pools_token0_id_"))
        batch_op.drop_index(batch_op.f("ix_pools_token1_id_"))

        batch_op.alter_column("token0_id_", new_column_name="token0_id")
        batch_op.alter_column("token1_id_", new_column_name="token1_id")


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



"""Uniswap V4: add exchange foreign key

Revision ID: 39e331854eb2
Revises: 901adb947000
Create Date: 2025-10-20 10:45:16.090760

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "39e331854eb2"
down_revision: str | Sequence[str] | None = "901adb947000"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""

    with op.batch_alter_table("pool_managers", schema=None) as batch_op:
        batch_op.add_column(sa.Column("exchange_id", sa.Integer(), nullable=False))
        batch_op.create_foreign_key("fk_exchange", "exchanges", ["exchange_id"], ["id"])


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



"""Copy token IDs to pool table

Revision ID: e453c9cd9e51
Revises: bbb8d61cef9c
Create Date: 2025-10-25 12:07:25.023957

"""

from collections.abc import Sequence
from typing import TYPE_CHECKING

import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import Session

from degenbot.database.models.pools import (
    AbstractUniswapV2Pool,
    AbstractUniswapV3Pool,
    AerodromeV2PoolTable,
    AerodromeV3PoolTable,
    CamelotV2PoolTable,
    LiquidityPoolTable,
    PancakeswapV2PoolTable,
    PancakeswapV3PoolTable,
    SushiswapV2PoolTable,
    SushiswapV3PoolTable,
    SwapbasedV2PoolTable,
    UniswapV2PoolTable,
    UniswapV3PoolTable,
)

# revision identifiers, used by Alembic.
revision: str = "e453c9cd9e51"
down_revision: str | Sequence[str] | None = "bbb8d61cef9c"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""

    connection = op.get_bind()
    session = Session(bind=connection)

    for table in [
        AerodromeV2PoolTable,
        AerodromeV3PoolTable,
        CamelotV2PoolTable,
        PancakeswapV2PoolTable,
        PancakeswapV3PoolTable,
        SushiswapV2PoolTable,
        SushiswapV3PoolTable,
        SwapbasedV2PoolTable,
        UniswapV2PoolTable,
        UniswapV3PoolTable,
    ]:
        if TYPE_CHECKING:
            assert isinstance(
                table,
                (AbstractUniswapV2Pool, AbstractUniswapV3Pool),
            )

        for pool in session.scalars(sa.select(table)).all():
            base_pool = session.scalar(
                sa.select(LiquidityPoolTable).where(LiquidityPoolTable.id == pool.pool_id)
            )

            assert isinstance(pool, (AbstractUniswapV2Pool, AbstractUniswapV3Pool))
            assert isinstance(base_pool, LiquidityPoolTable)

            base_pool.token0_id_ = pool.token0_id
            base_pool.token1_id_ = pool.token1_id

    session.commit()


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



"""Drop token0_id, token1_id from subclass tables

Revision ID: 082ee8a3d339
Revises: e453c9cd9e51
Create Date: 2025-10-25 13:41:16.069821

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "082ee8a3d339"
down_revision: str | Sequence[str] | None = "e453c9cd9e51"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""

    with op.batch_alter_table("aerodrome_v2_pools", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_aerodrome_v2_pools_token0_id"))
        batch_op.drop_index(batch_op.f("ix_aerodrome_v2_pools_token1_id"))
        batch_op.drop_column("token0_id")
        batch_op.drop_column("token1_id")

    with op.batch_alter_table("aerodrome_v3_pools", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_aerodrome_v3_pools_token0_id"))
        batch_op.drop_index(batch_op.f("ix_aerodrome_v3_pools_token1_id"))
        batch_op.drop_column("token0_id")
        batch_op.drop_column("token1_id")

    with op.batch_alter_table("camelot_v2_pools", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_camelot_v2_pools_token0_id"))
        batch_op.drop_index(batch_op.f("ix_camelot_v2_pools_token1_id"))
        batch_op.drop_column("token0_id")
        batch_op.drop_column("token1_id")

    with op.batch_alter_table("pancakeswap_v2_pools", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_pancakeswap_v2_pools_token0_id"))
        batch_op.drop_index(batch_op.f("ix_pancakeswap_v2_pools_token1_id"))
        batch_op.drop_column("token0_id")
        batch_op.drop_column("token1_id")

    with op.batch_alter_table("pancakeswap_v3_pools", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_pancakeswap_v3_pools_token0_id"))
        batch_op.drop_index(batch_op.f("ix_pancakeswap_v3_pools_token1_id"))
        batch_op.drop_column("token0_id")
        batch_op.drop_column("token1_id")

    with op.batch_alter_table("pools", schema=None) as batch_op:
        batch_op.alter_column("token0_id_", existing_type=sa.INTEGER(), nullable=False)
        batch_op.alter_column("token1_id_", existing_type=sa.INTEGER(), nullable=False)
        batch_op.create_index(batch_op.f("ix_pools_token0_id_"), ["token0_id_"], unique=False)
        batch_op.create_index(batch_op.f("ix_pools_token1_id_"), ["token1_id_"], unique=False)

    with op.batch_alter_table("sushiswap_v2_pools", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_sushiswap_v2_pools_token0_id"))
        batch_op.drop_index(batch_op.f("ix_sushiswap_v2_pools_token1_id"))
        batch_op.drop_column("token0_id")
        batch_op.drop_column("token1_id")

    with op.batch_alter_table("sushiswap_v3_pools", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_sushiswap_v3_pools_token0_id"))
        batch_op.drop_index(batch_op.f("ix_sushiswap_v3_pools_token1_id"))
        batch_op.drop_column("token0_id")
        batch_op.drop_column("token1_id")

    with op.batch_alter_table("swapbased_v2_pools", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_swapbased_v2_pools_token0_id"))
        batch_op.drop_index(batch_op.f("ix_swapbased_v2_pools_token1_id"))
        batch_op.drop_column("token0_id")
        batch_op.drop_column("token1_id")

    with op.batch_alter_table("uniswap_v2_pools", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_uniswap_v2_pools_token0_id"))
        batch_op.drop_index(batch_op.f("ix_uniswap_v2_pools_token1_id"))
        batch_op.drop_column("token0_id")
        batch_op.drop_column("token1_id")

    with op.batch_alter_table("uniswap_v3_pools", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_uniswap_v3_pools_token0_id"))
        batch_op.drop_index(batch_op.f("ix_uniswap_v3_pools_token1_id"))
        batch_op.drop_column("token0_id")
        batch_op.drop_column("token1_id")


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



import pathlib
import sqlite3

from alembic import command
from alembic.config import Config
from sqlalchemy import URL, create_engine, text
from sqlalchemy.orm import Session, scoped_session, sessionmaker

from degenbot.config import settings
from degenbot.database.models import Base
from degenbot.exceptions.database import BackupExists
from degenbot.logging import logger


def backup_sqlite_database(db_path: pathlib.Path) -> None:
    assert db_path.exists()

    backup_path = pathlib.Path(db_path).with_suffix(db_path.suffix + ".bak")
    if backup_path.exists():
        raise BackupExists(path=backup_path)

    engine = create_engine(
        f"sqlite:///{db_path.absolute()}",
    )
    with engine.connect() as connection:
        connection.execute(
            text("PRAGMA wal_checkpoint(FULL);"),
        )

    with sqlite3.connect(db_path) as src, sqlite3.connect(backup_path) as dest:
        src.backup(target=dest)


def create_new_sqlite_database(db_path: pathlib.Path) -> None:
    engine = create_engine(
        f"sqlite:///{db_path.absolute()}",
    )
    with engine.connect() as connection:
        assert (
            connection.execute(
                text("PRAGMA journal_mode=WAL;"),
            ).scalar()
            == "wal"
        )
        connection.execute(
            text("PRAGMA auto_vacuum=FULL;"),
        )

        Base.metadata.create_all(bind=engine)
        connection.execute(
            text("VACUUM;"),
        )

        logger.info(f"Initialized new SQLite database at {db_path}")
        command.stamp(get_alembic_config(), "head")


def compact_sqlite_database(db_path: pathlib.Path) -> None:
    engine = create_engine(
        f"sqlite:///{db_path.absolute()}",
    )
    with engine.connect() as connection:
        connection.execute(
            text("VACUUM;"),
        )
        logger.info(f"Compacted SQLite database at {db_path}")


def upgrade_existing_sqlite_database() -> None:
    command.upgrade(get_alembic_config(), "head")
    logger.info("Updated existing SQLite database.")


def get_scoped_sqlite_session(database_path: pathlib.Path) -> scoped_session[Session]:
    return scoped_session(
        session_factory=sessionmaker(
            bind=create_engine(
                URL.create(
                    drivername="sqlite",
                    database=str(database_path.absolute()),
                )
            )
        )
    )


def get_alembic_config() -> Config:
    cfg = Config()
    cfg.set_main_option("sqlalchemy.url", f"sqlite:///{settings.database.path.absolute()}")
    cfg.set_main_option("script_location", "degenbot:migrations")

    return cfg



from typing import Annotated

from sqlalchemy import ForeignKey, String
from sqlalchemy.orm import mapped_column

PrimaryKeyInt = Annotated[
    int,
    mapped_column(primary_key=True, autoincrement=True),
]
PrimaryForeignKeyPoolId = Annotated[
    int,
    mapped_column(ForeignKey("pools.id"), primary_key=True),
]
PrimaryForeignKeyManagedPoolId = Annotated[
    int,
    mapped_column(ForeignKey("managed_pools.id"), primary_key=True),
]
ForeignKeyManagedPoolId = Annotated[
    int,
    mapped_column(ForeignKey("managed_pools.id"), index=True),
]
ForeignKeyPoolId = Annotated[
    int,
    mapped_column(ForeignKey("pools.id"), index=True),
]
ForeignKeyPoolManagerId = Annotated[
    int,
    mapped_column(ForeignKey("pool_managers.id"), index=True),
]
ForeignKeyTokenId = Annotated[
    int,
    mapped_column(ForeignKey("erc20_tokens.id"), index=True),
]
ManagedPoolHash = Annotated[
    str,
    mapped_column(String(66)),
]



import tomllib
from pathlib import Path
from typing import Annotated

import tomlkit
from pydantic import BaseModel, HttpUrl, PlainSerializer, WebsocketUrl, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from degenbot.logging import logger
from degenbot.types.aliases import ChainId

CONFIG_DIR = Path.home() / ".config" / "degenbot"
CONFIG_FILE = CONFIG_DIR / "config.toml"
DB_PATH = CONFIG_DIR / "degenbot.db"


class DatabaseSettings(BaseModel):
    # Serialize the path as a string representation of the absolute path
    path: Annotated[
        Path,
        PlainSerializer(lambda path: str(path.absolute()), return_type=str),
    ]


class Settings(BaseSettings):
    model_config = SettingsConfigDict()

    database: DatabaseSettings
    rpc: dict[
        ChainId,
        HttpUrl | WebsocketUrl | Path,
    ]

    @field_validator("rpc", mode="after")
    def validate_paths(
        cls,  # noqa: N805
        rpc_dict: dict[ChainId, HttpUrl | WebsocketUrl | Path],
    ) -> dict[ChainId, HttpUrl | WebsocketUrl | Path]:
        """
        Validate the endpoints.

        This will convert all file paths to an absolute reference, leaving HTTP and WS URLs as-is.
        """

        return {
            chain_id: endpoint.expanduser().absolute() if isinstance(endpoint, Path) else endpoint
            for chain_id, endpoint in rpc_dict.items()
        }


def load_config_from_file(config_path: Path) -> Settings:
    return Settings.model_validate(
        tomllib.loads(
            config_path.read_text(),
        ),
    )


def save_config_to_file(config: Settings) -> None:
    CONFIG_FILE.write_text(
        tomlkit.dumps(
            config.model_dump(),
        ),
    )


if not CONFIG_DIR.exists():
    CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    logger.info(f"Created a configuration directory at {CONFIG_DIR}.")

if CONFIG_FILE.exists():
    settings = load_config_from_file(CONFIG_FILE)
else:
    settings = Settings(
        database=DatabaseSettings(
            path=DB_PATH,
        ),
        rpc={},
    )

    save_config_to_file(settings)
    logger.info(f"Created a configuration file at {CONFIG_FILE}.")

    if not settings.database.path.exists():
        from degenbot.database.operations import create_new_sqlite_database

        create_new_sqlite_database(db_path=settings.database.path)



"""Add unique index

Revision ID: d1f98e2c3b18
Revises: a4f59783919f
Create Date: 2025-09-05 13:53:55.488450

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "d1f98e2c3b18"
down_revision: str | Sequence[str] | None = "a4f59783919f"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.create_index(
        op.f("ix_uniswap_v4_pools_pool_hash"), "uniswap_v4_pools", ["pool_hash"], unique=True
    )


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



"""ERC-20: adjust index

Revision ID: 8aa4babb128a
Revises: 8c69198e6a21
Create Date: 2025-10-20 21:53:46.297738

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "8aa4babb128a"
down_revision: str | Sequence[str] | None = "8c69198e6a21"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""

    with op.batch_alter_table("erc20_tokens", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_erc20_tokens_chain_address"))
        batch_op.create_index("ix_erc20_tokens_address_chain", ["address", "chain"], unique=True)


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



"""Convert pool hash to text

Revision ID: 5c8805573ab3
Revises: 87fd9fc7ae00
Create Date: 2025-09-05 13:12:10.475697

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "5c8805573ab3"
down_revision: str | Sequence[str] | None = "87fd9fc7ae00"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.add_column(
        "uniswap_v4_pools", sa.Column("pool_hash_hex", sa.String(length=66), nullable=True)
    )
    op.execute("UPDATE uniswap_v4_pools SET pool_hash_hex = '0x' || hex(pool_hash)")
    with op.batch_alter_table("uniswap_v4_pools") as batch_op:
        batch_op.alter_column("pool_hash_hex", existing_type=sa.Text(), nullable=False)


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



from json import JSONDecodeError
from typing import TYPE_CHECKING, cast

import tenacity
from ujson import loads as ujson_loads
from web3 import AsyncBaseProvider, AsyncWeb3, JSONBaseProvider
from web3.types import RPCResponse

from degenbot.exceptions import DegenbotValueError
from degenbot.types.aliases import ChainId


def _fast_decode_rpc_response(raw_response: bytes) -> RPCResponse:
    """
    Decode the JSON-RPC response using ujson.
    """

    try:
        return cast("RPCResponse", ujson_loads(raw_response))
    except ValueError:
        # Re-raise as a dummy JSONDecodeError so web3py's exception handling works as intended.
        msg = "JSON failure"
        raise JSONDecodeError(msg, "[]", 0) from None


class AsyncConnectionManager:
    def __init__(self) -> None:
        self.connections: dict[ChainId, AsyncWeb3[AsyncBaseProvider]] = {}
        self._default_chain_id: ChainId | None = None

    def get_web3(self, chain_id: ChainId) -> AsyncWeb3[AsyncBaseProvider]:
        try:
            return self.connections[chain_id]
        except KeyError:
            raise DegenbotValueError(
                message="Chain ID does not have a registered Web3 instance."
            ) from None

    async def register_web3(
        self,
        w3: AsyncWeb3[AsyncBaseProvider],
        *,
        optimize: bool = True,
    ) -> None:
        async_w3_connected_check_with_retry = tenacity.AsyncRetrying(
            stop=tenacity.stop_after_delay(10),
            wait=tenacity.wait_exponential_jitter(),
            retry=tenacity.retry_if_result(lambda result: result is False),
        )
        try:
            await async_w3_connected_check_with_retry(w3.is_connected)
        except tenacity.RetryError as exc:
            raise DegenbotValueError(message="Web3 instance is not connected.") from exc

        if optimize:
            # Remove all middleware and monkey-patch the JSON decoding for RPC responses
            w3.middleware_onion.clear()
            if TYPE_CHECKING:
                assert isinstance(w3.provider, JSONBaseProvider)
            w3.provider.decode_rpc_response = _fast_decode_rpc_response

        self.connections[await w3.eth.chain_id] = w3

    def set_default_chain(self, chain_id: ChainId) -> None:
        self._default_chain_id = chain_id

    @property
    def default_chain_id(self) -> ChainId:
        if self._default_chain_id is None:
            raise DegenbotValueError(message="A default chain ID has not been provided.")
        return self._default_chain_id



from sqlalchemy import Index
from sqlalchemy.orm import Mapped

from .base import Address, Base
from .types import PrimaryKeyInt


class Erc20TokenTable(Base):
    __tablename__ = "erc20_tokens"

    id: Mapped[PrimaryKeyInt]
    chain: Mapped[int]
    address: Mapped[Address]
    name: Mapped[str | None]
    symbol: Mapped[str | None]
    decimals: Mapped[int | None]


# The (address, ChainId) tuple is unique for ERC-20 tokens
Index(
    "ix_erc20_tokens_address_chain",
    Erc20TokenTable.address,
    Erc20TokenTable.chain,
    unique=True,
)



"""add liquidity update marker

Revision ID: 7dc2ca38053f
Revises: 311beed36e7b
Create Date: 2025-10-14 14:49:00.008992

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "7dc2ca38053f"
down_revision: str | Sequence[str] | None = "311beed36e7b"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.add_column(
        "aerodrome_v3_pools", sa.Column("liquidity_update_block", sa.Integer(), nullable=True)
    )
    op.add_column(
        "aerodrome_v3_pools", sa.Column("liquidity_update_log_index", sa.Integer(), nullable=True)
    )
    op.add_column(
        "pancakeswap_v3_pools", sa.Column("liquidity_update_block", sa.Integer(), nullable=True)
    )
    op.add_column(
        "pancakeswap_v3_pools", sa.Column("liquidity_update_log_index", sa.Integer(), nullable=True)
    )
    op.add_column(
        "sushiswap_v3_pools", sa.Column("liquidity_update_block", sa.Integer(), nullable=True)
    )
    op.add_column(
        "sushiswap_v3_pools", sa.Column("liquidity_update_log_index", sa.Integer(), nullable=True)
    )
    op.add_column(
        "uniswap_v3_pools", sa.Column("liquidity_update_block", sa.Integer(), nullable=True)
    )
    op.add_column(
        "uniswap_v3_pools", sa.Column("liquidity_update_log_index", sa.Integer(), nullable=True)
    )


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



"""add stable column for Aerodrome V2

Revision ID: 723fe25c87b8
Revises: fec1e047973b
Create Date: 2025-10-01 11:37:11.820793

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "723fe25c87b8"
down_revision: str | Sequence[str] | None = "fec1e047973b"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.add_column("aerodrome_v2_pools", sa.Column("stable", sa.Boolean(), nullable=False))


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



from degenbot.constants import MAX_UINT256
from degenbot.exceptions.evm import EVMRevertError
from degenbot.uniswap.v4_libraries.functions import mulmod


def muldiv(
    a: int,
    b: int,
    denominator: int,
) -> int:
    """
    Calculates floor(a*b/denominator) with full precision. Throws if result overflows a uint256 or
    denominator == 0.

    ref: https://github.com/Uniswap/v4-core/blob/main/src/libraries/FullMath.sol

    Python integers do not overflow and have no bit depth limitation, so this function simply
    checks for an invalid result.
    """

    if denominator <= 0:
        msg = "required: denominator > 0"
        raise EVMRevertError(msg)

    result = (a * b) // denominator

    if result > MAX_UINT256:
        msg = "product > MAX_UINT256"
        raise EVMRevertError(msg)

    return result


def muldiv_rounding_up(
    a: int,
    b: int,
    denominator: int,
) -> int:
    """
    Calculates ceil(a*b//denominator) with full precision. Throws if result overflows a uint256 or
    denominator == 0.

    ref: https://github.com/Uniswap/v4-core/blob/main/src/libraries/FullMath.sol
    """

    if denominator <= 0:
        msg = "required: denominator > 0"
        raise EVMRevertError(msg)

    result = muldiv(a, b, denominator) + int(mulmod(a, b, denominator) > 0)

    if result > MAX_UINT256:
        msg = "product > MAX_UINT256"
        raise EVMRevertError(msg)

    return result



from sqlalchemy import ForeignKey, Index
from sqlalchemy.orm import Mapped, declared_attr, mapped_column, relationship

from degenbot.database.models.erc20 import Erc20TokenTable

from .base import Address, Base, BigInteger, ExchangeTable
from .types import (
    ForeignKeyManagedPoolId,
    ForeignKeyPoolId,
    ForeignKeyPoolManagerId,
    ForeignKeyTokenId,
    ManagedPoolHash,
    PrimaryForeignKeyManagedPoolId,
    PrimaryForeignKeyPoolId,
    PrimaryKeyInt,
)


class LiquidityPositionTable(Base):
    __tablename__ = "liquidity_positions"

    id: Mapped[PrimaryKeyInt]
    pool_id: Mapped[ForeignKeyPoolId]
    tick: Mapped[int]
    liquidity_net: Mapped[BigInteger]
    liquidity_gross: Mapped[BigInteger]


# The (PoolId, tick) tuple is unique for each liquidity position
Index(
    "ix_liquidity_positions_pool_id_tick",
    LiquidityPositionTable.pool_id,
    LiquidityPositionTable.tick,
    unique=True,
)


class InitializationMapTable(Base):
    __tablename__ = "initialization_maps"

    id: Mapped[PrimaryKeyInt]
    pool_id: Mapped[ForeignKeyPoolId]
    word: Mapped[int]
    bitmap: Mapped[BigInteger]


# The (PoolId, word) tuple is unique for each initialization map
Index(
    "ix_initialization_maps_pool_id_word",
    InitializationMapTable.pool_id,
    InitializationMapTable.word,
    unique=True,
)


class LiquidityPoolTable(Base):
    __tablename__ = "pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_on": "kind",
        "polymorphic_identity": "base",
    }

    id: Mapped[PrimaryKeyInt]
    address: Mapped[Address]
    chain: Mapped[int]
    kind: Mapped[str]

    token0_id: Mapped[int] = mapped_column(ForeignKey("erc20_tokens.id"), index=True)
    token1_id: Mapped[int] = mapped_column(ForeignKey("erc20_tokens.id"), index=True)

    token0: Mapped[Erc20TokenTable] = relationship("Erc20TokenTable", foreign_keys=token0_id)
    token1: Mapped[Erc20TokenTable] = relationship("Erc20TokenTable", foreign_keys=token1_id)

    exchange: Mapped[ExchangeTable] = relationship("ExchangeTable")
    exchange_id: Mapped[int] = mapped_column(ForeignKey("exchanges.id"))


# The (address, chainId) tuple is unique for each liquidity pool
Index(
    "ix_liquidity_pool_address_chain",
    LiquidityPoolTable.address,
    LiquidityPoolTable.chain,
    unique=True,
)


class UniswapFeeMixin:
    """
    A mixin class defining common columns for Uniswap V2 & V3 pools and variants.
    """

    fee_token0: Mapped[int]
    fee_token1: Mapped[int]
    fee_denominator: Mapped[int]


class AbstractUniswapV2Pool(LiquidityPoolTable, UniswapFeeMixin):
    """
    This abstract class represents a common parent for all Uniswap V2 variants. It may be used to
    identify concrete subclasses at runtime, but otherwise is not useful for performing database
    queries.
    """

    __abstract__ = True


class AerodromeV2PoolTable(AbstractUniswapV2Pool):
    __tablename__ = "aerodrome_v2_pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "aerodrome_v2",
    }

    stable: Mapped[bool]
    pool_id: Mapped[PrimaryForeignKeyPoolId]


class CamelotV2PoolTable(AbstractUniswapV2Pool):
    __tablename__ = "camelot_v2_pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "camelot_v2",
    }

    pool_id: Mapped[PrimaryForeignKeyPoolId]


class PancakeswapV2PoolTable(AbstractUniswapV2Pool):
    __tablename__ = "pancakeswap_v2_pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "pancakeswap_v2",
    }

    pool_id: Mapped[PrimaryForeignKeyPoolId]


class SushiswapV2PoolTable(AbstractUniswapV2Pool):
    __tablename__ = "sushiswap_v2_pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "sushiswap_v2",
    }

    pool_id: Mapped[PrimaryForeignKeyPoolId]


class SwapbasedV2PoolTable(AbstractUniswapV2Pool):
    __tablename__ = "swapbased_v2_pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "swapbased_v2",
    }

    pool_id: Mapped[PrimaryForeignKeyPoolId]


class UniswapV2PoolTable(AbstractUniswapV2Pool):
    __tablename__ = "uniswap_v2_pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "uniswap_v2",
    }

    pool_id: Mapped[PrimaryForeignKeyPoolId]


class AbstractUniswapV3Pool(LiquidityPoolTable, UniswapFeeMixin):
    """
    This abstract class represents a parent for all Uniswap V3 variants. It may be used to identify
    concrete subclasses at runtime.
    """

    __abstract__ = True

    tick_spacing: Mapped[int]

    liquidity_update_block: Mapped[int | None]
    liquidity_update_log_index: Mapped[int | None]

    @declared_attr
    @classmethod
    def liquidity_positions(cls) -> Mapped[list[LiquidityPositionTable]]:
        return relationship(
            "LiquidityPositionTable",
            cascade="all, delete",
        )

    @declared_attr
    @classmethod
    def initialization_maps(cls) -> Mapped[list[InitializationMapTable]]:
        return relationship(
            "InitializationMapTable",
            cascade="all, delete",
        )


class AerodromeV3PoolTable(AbstractUniswapV3Pool):
    __tablename__ = "aerodrome_v3_pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "aerodrome_v3",
    }

    pool_id: Mapped[PrimaryForeignKeyPoolId]


class UniswapV3PoolTable(AbstractUniswapV3Pool):
    __tablename__ = "uniswap_v3_pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "uniswap_v3",
    }

    pool_id: Mapped[PrimaryForeignKeyPoolId]


class PancakeswapV3PoolTable(AbstractUniswapV3Pool):
    __tablename__ = "pancakeswap_v3_pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "pancakeswap_v3",
    }

    pool_id: Mapped[PrimaryForeignKeyPoolId]


class SushiswapV3PoolTable(AbstractUniswapV3Pool):
    __tablename__ = "sushiswap_v3_pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "sushiswap_v3",
    }

    pool_id: Mapped[PrimaryForeignKeyPoolId]


class ManagedPoolLiquidityPositionTable(Base):
    __tablename__ = "managed_pool_liquidity_positions"

    id: Mapped[PrimaryKeyInt]
    managed_pool_id: Mapped[ForeignKeyManagedPoolId]
    tick: Mapped[int]
    liquidity_net: Mapped[BigInteger]
    liquidity_gross: Mapped[BigInteger]


# The (PoolId, tick) tuple is unique for each liquidity position
Index(
    "ix_managed_pool_liquidity_positions_pool_id_tick",
    ManagedPoolLiquidityPositionTable.managed_pool_id,
    ManagedPoolLiquidityPositionTable.tick,
    unique=True,
)


class ManagedPoolInitializationMapTable(Base):
    __tablename__ = "managed_pool_initialization_maps"

    id: Mapped[PrimaryKeyInt]
    managed_pool_id: Mapped[ForeignKeyManagedPoolId]
    word: Mapped[int]
    bitmap: Mapped[BigInteger]


# The (ManagedPoolId, word) tuple is unique for each initialization map
Index(
    "ix_managed_pool_initialization_maps_pool_id_word",
    ManagedPoolInitializationMapTable.managed_pool_id,
    ManagedPoolInitializationMapTable.word,
    unique=True,
)


class PoolManagerTable(Base):
    __tablename__ = "pool_managers"

    id: Mapped[PrimaryKeyInt]
    address: Mapped[Address]
    chain: Mapped[int]
    kind: Mapped[str]

    exchange_id: Mapped[int] = mapped_column(ForeignKey("exchanges.id"))
    exchange: Mapped[ExchangeTable] = relationship("ExchangeTable")


# The (address, chainId) tuple is unique for each pool manager
Index(
    "ix_pool_manager_address_chain",
    PoolManagerTable.address,
    PoolManagerTable.chain,
    unique=True,
)


class ManagedLiquidityPoolTable(Base):
    __tablename__ = "managed_pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_on": "kind",
        "polymorphic_identity": "base",
    }

    id: Mapped[PrimaryKeyInt]
    kind: Mapped[str]

    manager_id: Mapped[ForeignKeyPoolManagerId]
    manager: Mapped[PoolManagerTable] = relationship("PoolManagerTable")


class AbstractUniswapV4Pool(ManagedLiquidityPoolTable):
    """
    This abstract class represents a parent for all Uniswap V4 variants. It should not be
    instantiated directly, but may be used to query and select child classes.
    """

    __abstract__ = True

    managed_pool_id: Mapped[PrimaryForeignKeyManagedPoolId]

    pool_hash: Mapped[ManagedPoolHash] = mapped_column(index=True, unique=True)
    hooks: Mapped[Address]
    currency0_id: Mapped[ForeignKeyTokenId]
    currency1_id: Mapped[ForeignKeyTokenId]
    fee_currency0: Mapped[int]
    fee_currency1: Mapped[int]
    fee_denominator: Mapped[int]
    tick_spacing: Mapped[int]

    liquidity_update_block: Mapped[int | None]
    liquidity_update_log_index: Mapped[int | None]

    @declared_attr
    @classmethod
    def liquidity_positions(cls) -> Mapped[list[ManagedPoolLiquidityPositionTable]]:
        return relationship(
            "ManagedPoolLiquidityPositionTable",
            cascade="all, delete",
        )

    @declared_attr
    @classmethod
    def initialization_maps(cls) -> Mapped[list[ManagedPoolInitializationMapTable]]:
        return relationship(
            "ManagedPoolInitializationMapTable",
            cascade="all, delete",
        )

    @declared_attr
    @classmethod
    def currency0(cls) -> Mapped[Erc20TokenTable]:
        return relationship(
            "Erc20TokenTable",
            foreign_keys=cls.currency0_id,
        )

    @declared_attr
    @classmethod
    def currency1(cls) -> Mapped[Erc20TokenTable]:
        return relationship(
            "Erc20TokenTable",
            foreign_keys=cls.currency1_id,
        )


class UniswapV4PoolTable(AbstractUniswapV4Pool):
    __tablename__ = "uniswap_v4_pools"
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_on": "kind",
        "polymorphic_identity": "uniswap_v4",
    }



# This module is adapted from the Uniswap V4 BitMath.sol library.
# Reference: https://github.com/Uniswap/v4-core/blob/main/src/libraries/BitMath.sol


def least_significant_bit(number: int) -> int:
    """
    Find the least significant bit for the given number.
    """

    if number <= 0:
        msg = "Number must be >0"
        raise ValueError(msg)

    return (number & -number).bit_length() - 1


def least_significant_bit_legacy(number: int) -> int:
    """
    Find the least significant bit for the given number.
    """

    if number <= 0:
        msg = "Number must be >0"
        raise ValueError(msg)

    for index in range(256):
        if (number >> index) & 1 == 1:
            return index

    raise ValueError  # should be unreachable for valid 256 bit numbers


def most_significant_bit(number: int) -> int:
    """
    Find the most significant bit for the given number.
    """

    if number <= 0:
        msg = "Number must be >0"
        raise ValueError(msg)

    return number.bit_length() - 1


def most_significant_bit_legacy(number: int) -> int:
    """
    Find the most significant bit for the given number.
    """

    if number <= 0:
        msg = "Number must be >0"
        raise ValueError(msg)

    for index in range(256):
        if (number >> index) == 1:
            return index

    raise ValueError  # should be unreachable for valid 256 bit numbers



from .base import Address, Base, BigInteger, IntMappedToString
from .erc20 import Erc20TokenTable
from .pools import (
    AerodromeV2PoolTable,
    AerodromeV3PoolTable,
    CamelotV2PoolTable,
    InitializationMapTable,
    LiquidityPoolTable,
    LiquidityPositionTable,
    ManagedLiquidityPoolTable,
    ManagedPoolInitializationMapTable,
    ManagedPoolLiquidityPositionTable,
    PancakeswapV2PoolTable,
    PancakeswapV3PoolTable,
    PoolManagerTable,
    SushiswapV2PoolTable,
    SushiswapV3PoolTable,
    SwapbasedV2PoolTable,
    UniswapV2PoolTable,
    UniswapV3PoolTable,
    UniswapV4PoolTable,
)

__all__ = (
    "Address",
    "AerodromeV2PoolTable",
    "AerodromeV3PoolTable",
    "Base",
    "BigInteger",
    "CamelotV2PoolTable",
    "Erc20TokenTable",
    "InitializationMapTable",
    "IntMappedToString",
    "LiquidityPoolTable",
    "LiquidityPositionTable",
    "ManagedLiquidityPoolTable",
    "ManagedPoolInitializationMapTable",
    "ManagedPoolLiquidityPositionTable",
    "PancakeswapV2PoolTable",
    "PancakeswapV3PoolTable",
    "PoolManagerTable",
    "SushiswapV2PoolTable",
    "SushiswapV3PoolTable",
    "SwapbasedV2PoolTable",
    "UniswapV2PoolTable",
    "UniswapV3PoolTable",
    "UniswapV4PoolTable",
)



"""Update Uniswap V4 table indices

Revision ID: 87fd9fc7ae00
Revises: 756fba1f75f4
Create Date: 2025-09-05 10:37:01.739730

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "87fd9fc7ae00"
down_revision: str | Sequence[str] | None = "756fba1f75f4"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.drop_index(op.f("ix_pool_managers_address"), table_name="pool_managers")
    op.create_index(op.f("ix_pool_managers_address"), "pool_managers", ["address"], unique=True)
    op.drop_index(op.f("ix_managed_pool_hash"), table_name="uniswap_v4_pools")
    op.create_index(
        op.f("ix_uniswap_v4_pools_pool_hash"), "uniswap_v4_pools", ["pool_hash"], unique=True
    )


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



"""V4 pools: add token foreign key

Revision ID: 8b29ed99a2ac
Revises: 50b39bafa0be
Create Date: 2025-10-17 14:33:55.170773

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

from degenbot.database.models.pools import UniswapV4PoolTable

# revision identifiers, used by Alembic.
revision: str = "8b29ed99a2ac"
down_revision: str | Sequence[str] | None = "50b39bafa0be"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""

    with op.batch_alter_table(UniswapV4PoolTable.__tablename__, schema=None) as batch_op:
        batch_op.add_column(sa.Column("currency0_id", sa.Integer(), nullable=False))
        batch_op.add_column(sa.Column("currency1_id", sa.Integer(), nullable=False))
        batch_op.create_foreign_key("fk_currency0_id", "erc20_tokens", ["currency0_id"], ["id"])
        batch_op.create_foreign_key("fk_currency1_id", "erc20_tokens", ["currency1_id"], ["id"])

        # Drop the token address columns
        batch_op.drop_column("currency0")
        batch_op.drop_column("currency1")


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



from alembic.runtime.migration import MigrationContext
from alembic.script import ScriptDirectory

from degenbot.config import settings
from degenbot.database.operations import get_alembic_config, get_scoped_sqlite_session
from degenbot.logging import logger
from degenbot.version import __version__

db_session = get_scoped_sqlite_session(database_path=settings.database.path)

current_database_version = MigrationContext.configure(
    connection=db_session().connection()
).get_current_revision()
latest_database_version = ScriptDirectory.from_config(
    config=get_alembic_config()
).get_current_head()


if current_database_version is not None and current_database_version != latest_database_version:
    logger.warning(
        f"The current database revision ({current_database_version}) does not match the latest "
        f"({latest_database_version}) for {__package__} version {__version__}!"
        "\n"
        "Database-related features may raise exceptions if you continue. Perform database "
        "migrations with 'degenbot database upgrade'."
    )



from . import (
    abi as abi,
)  # excluded from __all__ so it doesn't bubble back up to the top level package namespace
from .managers import PancakeV2PoolManager, PancakeV3PoolManager
from .pools import PancakeV2Pool, PancakeV3Pool

__all__ = (
    "PancakeV2Pool",
    "PancakeV2PoolManager",
    "PancakeV3Pool",
    "PancakeV3PoolManager",
)



"""Add indexes for non-primary foreign keys

Revision ID: 03e723f439dc
Revises: b20f5564b3b6
Create Date: 2025-10-24 23:16:27.369363

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "03e723f439dc"
down_revision: str | Sequence[str] | None = "b20f5564b3b6"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""

    with op.batch_alter_table("aerodrome_v2_pools", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_aerodrome_v2_pools_token0_id"), ["token0_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_aerodrome_v2_pools_token1_id"), ["token1_id"], unique=False
        )

    with op.batch_alter_table("aerodrome_v3_pools", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_aerodrome_v3_pools_token0_id"), ["token0_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_aerodrome_v3_pools_token1_id"), ["token1_id"], unique=False
        )

    with op.batch_alter_table("camelot_v2_pools", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_camelot_v2_pools_token0_id"), ["token0_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_camelot_v2_pools_token1_id"), ["token1_id"], unique=False
        )

    with op.batch_alter_table("initialization_maps", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_initialization_maps_pool_id"), ["pool_id"], unique=False
        )

    with op.batch_alter_table("liquidity_positions", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_liquidity_positions_pool_id"), ["pool_id"], unique=False
        )

    with op.batch_alter_table("managed_pool_initialization_maps", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_managed_pool_initialization_maps_managed_pool_id"),
            ["managed_pool_id"],
            unique=False,
        )

    with op.batch_alter_table("managed_pool_liquidity_positions", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_managed_pool_liquidity_positions_managed_pool_id"),
            ["managed_pool_id"],
            unique=False,
        )

    with op.batch_alter_table("managed_pools", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_managed_pools_manager_id"), ["manager_id"], unique=False
        )

    with op.batch_alter_table("pancakeswap_v2_pools", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_pancakeswap_v2_pools_token0_id"), ["token0_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_pancakeswap_v2_pools_token1_id"), ["token1_id"], unique=False
        )

    with op.batch_alter_table("pancakeswap_v3_pools", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_pancakeswap_v3_pools_token0_id"), ["token0_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_pancakeswap_v3_pools_token1_id"), ["token1_id"], unique=False
        )

    with op.batch_alter_table("sushiswap_v2_pools", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_sushiswap_v2_pools_token0_id"), ["token0_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_sushiswap_v2_pools_token1_id"), ["token1_id"], unique=False
        )

    with op.batch_alter_table("sushiswap_v3_pools", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_sushiswap_v3_pools_token0_id"), ["token0_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_sushiswap_v3_pools_token1_id"), ["token1_id"], unique=False
        )

    with op.batch_alter_table("swapbased_v2_pools", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_swapbased_v2_pools_token0_id"), ["token0_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_swapbased_v2_pools_token1_id"), ["token1_id"], unique=False
        )

    with op.batch_alter_table("uniswap_v2_pools", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_uniswap_v2_pools_token0_id"), ["token0_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_uniswap_v2_pools_token1_id"), ["token1_id"], unique=False
        )

    with op.batch_alter_table("uniswap_v3_pools", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_uniswap_v3_pools_token0_id"), ["token0_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_uniswap_v3_pools_token1_id"), ["token1_id"], unique=False
        )

    with op.batch_alter_table("uniswap_v4_pools", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_uniswap_v4_pools_currency0_id"), ["currency0_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_uniswap_v4_pools_currency1_id"), ["currency1_id"], unique=False
        )


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



"""add factory and deployer to exchange table

Revision ID: 311beed36e7b
Revises: 723fe25c87b8
Create Date: 2025-10-03 15:27:06.651242

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "311beed36e7b"
down_revision: str | Sequence[str] | None = "723fe25c87b8"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.alter_column("exchanges", "factory", existing_type=sa.VARCHAR(length=42), nullable=False)


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



"""V2 pools: convert token0/token1 to foreign keys

Revision ID: b8e0b921299a
Revises: fb9abf7c7d57
Create Date: 2025-10-17 11:21:25.547909

"""

from collections.abc import Sequence
from typing import TYPE_CHECKING

import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import Session

from degenbot.database.models.erc20 import Erc20TokenTable
from degenbot.database.models.pools import (
    AbstractUniswapV2Pool,
    AerodromeV2PoolTable,
    CamelotV2PoolTable,
    PancakeswapV2PoolTable,
    SushiswapV2PoolTable,
    SwapbasedV2PoolTable,
    UniswapV2PoolTable,
)

# revision identifiers, used by Alembic.
revision: str = "b8e0b921299a"
down_revision: str | Sequence[str] | None = "fb9abf7c7d57"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""

    connection = op.get_bind()
    session = Session(bind=connection)

    # Build all token references
    tokens: set[tuple[int, str]] = set()

    for table in [
        AerodromeV2PoolTable,
        CamelotV2PoolTable,
        PancakeswapV2PoolTable,
        SushiswapV2PoolTable,
        SwapbasedV2PoolTable,
        UniswapV2PoolTable,
    ]:
        if TYPE_CHECKING:
            assert isinstance(table, AbstractUniswapV2Pool)

        tokens.update(session.query(table.chain, table.token0).all())
        tokens.update(session.query(table.chain, table.token1).all())

        token_ids: dict[tuple[int, str], int] = {
            (chain_id, token_address): session.scalar(
                sa.select(Erc20TokenTable.id).where(
                    Erc20TokenTable.chain == chain_id,
                    Erc20TokenTable.address == token_address,
                )
            )
            for chain_id, token_address in tokens
        }

        with op.batch_alter_table(table.__tablename__, schema=None) as batch_op:
            batch_op.add_column(sa.Column("token0_id", sa.Integer(), nullable=True))
            batch_op.add_column(sa.Column("token1_id", sa.Integer(), nullable=True))
            batch_op.create_foreign_key("fk_token0_id", "erc20_tokens", ["token0_id"], ["id"])
            batch_op.create_foreign_key("fk_token1_id", "erc20_tokens", ["token1_id"], ["id"])

        for pool_id, token0, token1 in connection.execute(
            sa.text(f"SELECT pool_id,token0,token1 FROM {table.__tablename__}")  # noqa: S608
        ).fetchall():
            pool_chain = session.scalar(sa.select(table.chain).where(table.pool_id == pool_id))
            connection.execute(
                sa.text(
                    f"""
                    UPDATE {table.__tablename__}\
                    SET token0_id = :token0_value, token1_id = :token1_value\
                    WHERE pool_id = :pool_id"""  # noqa: S608
                ),
                {
                    "pool_id": pool_id,
                    "token0_value": token_ids[(pool_chain, token0)],
                    "token1_value": token_ids[(pool_chain, token1)],
                },
            )

        with op.batch_alter_table(table.__tablename__, schema=None) as batch_op:
            # Drop the token address columns
            batch_op.drop_column("token0")
            batch_op.drop_column("token1")

            # Make column non-nullable after the keys are populated
            batch_op.alter_column("token0_id", existing_type=sa.Integer(), nullable=False)
            batch_op.alter_column("token1_id", existing_type=sa.Integer(), nullable=False)


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



import click

from degenbot.cli import cli
from degenbot.config import settings
from degenbot.database import current_database_version, latest_database_version
from degenbot.database.operations import (
    backup_sqlite_database,
    compact_sqlite_database,
    create_new_sqlite_database,
    upgrade_existing_sqlite_database,
)
from degenbot.exceptions.database import BackupExists
from degenbot.version import __version__


@cli.group()
def database() -> None:
    """
    Database commands
    """


@database.command("backup")
def database_backup() -> None:
    """
    Back up the database.
    """

    try:
        backup_sqlite_database(settings.database.path)
    except BackupExists as exc:
        user_confirm = click.confirm(
            f"An existing backup was found at {exc.path}. Do you want to remove it and continue?",
            default=False,
        )
        if user_confirm:
            exc.path.unlink()
            backup_sqlite_database(settings.database.path)
        else:
            raise click.Abort from None


@database.command("reset")
def database_reset() -> None:
    """
    Remove and recreate the database.
    """

    user_confirm = click.confirm(
        f"The existing database at {settings.database.path} will be removed and a new, empty database will be created and initialized using the schema included in {__package__} version {__version__}. Do you want to proceed?",  # noqa: E501
        default=False,
    )
    if user_confirm:
        settings.database.path.unlink()
        create_new_sqlite_database(settings.database.path)
    else:
        raise click.Abort


@database.command("upgrade")
@click.option(
    "--force",
    is_flag=True,
    help="Skip confirmation prompt",
)
def database_upgrade(*, force: bool) -> None:
    """
    Upgrade the database to the latest schema.
    """

    if force or click.confirm(
        f"The database at {settings.database.path} will be upgraded from version {current_database_version} to {latest_database_version}. Do you want to proceed?",  # noqa:E501
        default=False,
    ):
        upgrade_existing_sqlite_database()
    else:
        raise click.Abort


@database.command("compact")
def database_compact() -> None:
    """
    Compact the database.
    """
    compact_sqlite_database(settings.database.path)



"""Drop binary pool hash column

Revision ID: 3199199def8c
Revises: 5c8805573ab3
Create Date: 2025-09-05 13:38:26.500135

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "3199199def8c"
down_revision: str | Sequence[str] | None = "5c8805573ab3"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.drop_index(op.f("ix_uniswap_v4_pools_pool_hash"), table_name="uniswap_v4_pools")
    with op.batch_alter_table("uniswap_v4_pools") as batch_op:
        batch_op.drop_column("pool_hash")


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



"""Convert index

Revision ID: fec1e047973b
Revises: 3eb6b82dc42d
Create Date: 2025-09-30 14:53:06.171358

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "fec1e047973b"
down_revision: str | Sequence[str] | None = "3eb6b82dc42d"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.drop_index(op.f("ix_pools_address"), table_name="pools")
    op.create_index("ix_liquidity_pool_address_chain", "pools", ["address", "chain"], unique=True)


def downgrade() -> None:
    """Downgrade schema."""
    msg = "Downgrade is not supported for this migration."
    raise NotImplementedError(msg)



import pathlib

from degenbot.exceptions.base import DegenbotError


class BackupExists(DegenbotError):
    """
    The rate of exchange for the path is below the minimum.
    """

    def __init__(self, path: pathlib.Path) -> None:
        self.path = path
        super().__init__(message=f"A backup at {path} already exists.")
